#
# CDDL HEADER START
#
# The contents of this file are subject to the terms of the
# Common Development and Distribution License (the "License").
# You may not use this file except in compliance with the License.
#
# You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
# or http://www.opensolaris.org/os/licensing.
# See the License for the specific language governing permissions
# and limitations under the License.
#
# When distributing Covered Code, include this CDDL HEADER in each
# file and include the License file at usr/src/OPENSOLARIS.LICENSE.
# If applicable, add the following below this CDDL HEADER, with the
# fields enclosed by brackets "[]" replaced with your own identifying
# information: Portions Copyright [yyyy] [name of copyright owner]
#
# CDDL HEADER END
#

#
# Copyright (c) 2011, Oracle and/or its affiliates. All rights reserved.
#

#
# Written by Roland Mainz <roland.mainz@nrubsig.org>
#

#
# This module tests whether making IFS a static variable works
#
# This was reported as CR #6990608 ("ksh93 crash with static IFS"):
# ------------ snip ------------
# Description
#     The test case here crashes /bin/ksh93:
# =================================
# function printarray
# {
# 	typeset mode=$1
# 	nameref ar=$2
# 	typeset -S IFS # static variable
# 	
# 	if [[ $mode == 'set' ]] ; then
# 		IFS=';'
# 	elif [[ $mode == 'unset' ]] ; then
# 		unset IFS
# 	fi
# 	
# 	print -- "${ar[*]}"
# }
# function run
# {
# 	typeset -a data=(
# 		'aaa'
# 		'bbb'
# 		'ccc'
# 		'ddd'
# 	)
# 	
# 	printarray set data
# 	printarray print data
# 	printarray unset data
# 	print -- "${data[*]}"
# 	printarray set data
# 	printarray print data
# 	printarray unset data
# 	print -- "${data[*]}"
# }
# run
# =================================
# gdb is less than useful with optimized code:
# (gdb) bt
# #0  0x00000001001199f0 in ?? ()
# #1  0xffffffff7d5535ec in sh_macexpand () from /usr/lib/64/libshell.so.1
# #2  0xffffffff7d538370 in sh_argbuild () from /usr/lib/64/libshell.so.1
# #3  0xffffffff7d582488 in sh_exec () from /usr/lib/64/libshell.so.1
# #4  0xffffffff7d582404 in sh_exec () from /usr/lib/64/libshell.so.1
# #5  0xffffffff7d588894 in sh_funscope () from /usr/lib/64/libshell.so.1
# #6  0xffffffff7d588cfc in sh_funct () from /usr/lib/64/libshell.so.1
# #7  0xffffffff7d583ea4 in sh_exec () from /usr/lib/64/libshell.so.1
# #8  0xffffffff7d5823d8 in sh_exec () from /usr/lib/64/libshell.so.1
# #9  0xffffffff7d588894 in sh_funscope () from /usr/lib/64/libshell.so.1
# #10 0xffffffff7d588cfc in sh_funct () from /usr/lib/64/libshell.so.1
# #11 0xffffffff7d583ea4 in sh_exec () from /usr/lib/64/libshell.so.1
# #12 0xffffffff7d55c87c in exfile () from /usr/lib/64/libshell.so.1
# #13 0xffffffff7d55bc44 in sh_main () from /usr/lib/64/libshell.so.1
# #14 0x0000000100000d60 in main ()
# Frequency
#     Always
# Regression
#     no
# Steps to Reproduce
#     Run testcase
# Expected Result
#     aaa;bbb;ccc;ddd
# aaa;bbb;ccc;ddd
# aaa bbb ccc ddd
# aaa bbb ccc ddd
# aaa;bbb;ccc;ddd
# aaa;bbb;ccc;ddd
# aaa bbb ccc ddd
# aaa bbb ccc ddd
# Actual Result
#     crash, coredump
# Error Message(s)
#     coredump
# Test Case
#     function printarray
# {
# 	typeset mode=$1
# 	nameref ar=$2
# 	typeset -S IFS # static variable
# 	
# 	if [[ $mode == 'set' ]] ; then
# 		IFS=';'
# 	elif [[ $mode == 'unset' ]] ; then
# 		unset IFS
# 	fi
# 	
# 	print -- "${ar[*]}"
# }
# function run
# {
# 	typeset -a data=(
# 		'aaa'
# 		'bbb'
# 		'ccc'
# 		'ddd'
# 	)
# 	
# 	printarray set data
# 	printarray print data
# 	printarray unset data
# 	print -- "${data[*]}"
# 	printarray set data
# 	printarray print data
# 	printarray unset data
# 	print -- "${data[*]}"
# }
# run
# Workaround
#     Make typeset -S a global variable
# Additional configuration information
#     Solaris 11, tosiba laptop
# ------------ snip ------------
#

# test setup
function err_exit
{
	print -u2 -n '\t'
	print -u2 -r ${Command}[$1]: "${@:2}"
	(( Errors++ ))
}
alias err_exit='err_exit $LINENO'

set -o nounset
Command=${0##*/}
integer Errors=0

typeset ocwd
typeset tmpdir

# create temporary test directory
ocwd="${PWD}"
tmpdir="${ mktemp -t -d 'test_sun_solaris_cr_6990608_static_ifs_crash.XXXXXXXX' ; }" || err_exit 'Cannot create temporary directory'

cd "${tmpdir}" || { err_exit "cd ${tmpdir} failed." ; exit $((Errors<125?Errors:125)) ; }

function test1
{
	compound out=( typeset stdout stderr ; integer res )
	typeset testname="${0}"

	cat >'testscript1.sh' <<EOF
function printarray
{
	typeset mode="\$1"
	nameref ar=\$2
	typeset -S IFS
	
	if [[ \$mode == 'set' ]] ; then
		IFS=';'
	elif [[ \$mode == 'unset' ]] ; then
		unset IFS
	fi
	
	print -- "\${ar[*]}"
	return 0
}


function main
{
	typeset -a data=(
		'aaa'
		'bbb'
		'ccc'
		'ddd'
	)
	
	printarray set data
	printarray print data
	printarray unset data
	print -- "\${data[*]}"

	printarray set data
	printarray print data
	printarray unset data
	print -- "\${data[*]}"
	
	return 0
}

main "\$@"
EOF

	typeset -r expected_output=$'aaa;bbb;ccc;ddd\naaa;bbb;ccc;ddd\naaa bbb ccc ddd\naaa bbb ccc ddd\naaa;bbb;ccc;ddd\naaa;bbb;ccc;ddd\naaa bbb ccc ddd\naaa bbb ccc ddd'
	compound out=( typeset stdout stderr ; integer res )
	typeset testname="${0}"
	typeset s=''
	integer i
	integer saved_Errors
	typeset testscript
	
	# provide a compiled version of the script
	${SHCOMP}    'testscript1.sh' 'testscript1.shbin'
	${SHCOMP} -n 'testscript1.sh' 'testscript1_n.shbin'
	
	for testscript in \
		'testscript1.sh' \
		'testscript1.shbin' \
		'testscript1_n.shbin' ; do
		(( saved_Errors=Errors ))

		# The crash depends on a precise memory layout and allocation/deallocation
		# sequence. To ensure we trigger the error we pass the shell an environment
		# variable called "s" with a ever-growing (per iteration) string length
		for (( i=2 ; i < 1024 && (Errors-saved_Errors) < 2 ; i++ )) ; do
			out.stderr="${ { out.stdout="${ s="${s}${i}" ; s="$s" ${SHELL} -o nounset "${testscript}" ; (( out.res=$? )) ; }" ; } 2>&1 ; }"

			[[ "${out.stdout}" == "${expected_output}" ]] || err_exit "${testname}: Expected stdout==${ printf '%q\n' "${expected_output}" ;}, got ${ printf '%q\n' "${out.stdout}" ; }"
			[[ "${out.stderr}" == '' ]] || err_exit "${testname}: Expected empty stderr, got ${ printf '%q\n' "${out.stderr}" ; }"
			(( out.res == 0 )) || err_exit "${testname}: Unexpected exit code ${out.res}"
		done
	done

	rm 'testscript1.sh' 'testscript1.shbin' 'testscript1_n.shbin'

	return 0
}


# run tests
test1


cd "${ocwd}"
rmdir "${tmpdir}" || err_exit "Cannot remove temporary directory ${tmpdir}."

# tests done
exit $((Errors<125?Errors:125))
