#
# CDDL HEADER START
#
# The contents of this file are subject to the terms of the
# Common Development and Distribution License (the "License").
# You may not use this file except in compliance with the License.
#
# You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
# or http://www.opensolaris.org/os/licensing.
# See the License for the specific language governing permissions
# and limitations under the License.
#
# When distributing Covered Code, include this CDDL HEADER in each
# file and include the License file at usr/src/OPENSOLARIS.LICENSE.
# If applicable, add the following below this CDDL HEADER, with the
# fields enclosed by brackets "[]" replaced with your own identifying
# information: Portions Copyright [yyyy] [name of copyright owner]
#
# CDDL HEADER END
#

#
# Copyright (c) 2011, Oracle and/or its affiliates. All rights reserved.
#

#
# Written by Roland Mainz <roland.mainz@nrubsig.org>
#

#
# This test module contains misc math tests
#

# test setup
function err_exit
{
	print -u2 -n '\t'
	print -u2 -r ${Command}[$1]: "${@:2}"
	(( Errors++ ))
}
alias err_exit='err_exit $LINENO'

set -o nounset
Command=${0##*/}
integer Errors=0

set -o nounset

typeset ocwd
typeset tmpdir

# create temporary test directory
ocwd="$PWD"
tmpdir="$(mktemp -t -d 'test_sun_solaris_math_misc.XXXXXXXX')" || err_exit 'Cannot create temporary directory'

cd "${tmpdir}" || { err_exit "cd ${tmpdir} failed." ; exit $((Errors<125?Errors:125)) ; }


function test_arithmetric_expression_accesss_array_element_through_nameref
{
        compound out=( typeset stdout stderr ; integer res )
	compound -r -a tests=(
		(
			cmd='@@TYPE@@ -a @@VAR@@ ;  @@VAR@@[1]=90 ;       function x { nameref nz=$1 ;              print " $(( round(nz) ))==$(( round($nz) ))" ; } ; x @@VAR@@[1]'		; stdoutpattern=' 90==90'
		)
		(
			cmd='@@TYPE@@ -a @@VAR@@=( [1]=90 ) ;             function x { nameref nz=$1 ;              print " $(( round(nz) ))==$(( round($nz) ))" ; } ; x @@VAR@@[1]'		; stdoutpattern=' 90==90'
		)
		(
			cmd='@@TYPE@@ -a @@VAR@@ ;  @@VAR@@[1][3]=90 ;    function x { nameref nz=$1 ;               print " $(( round(nz) ))==$(( round($nz) ))" ; } ; x @@VAR@@[1][3]'	; stdoutpattern=' 90==90'
		)
		# This crashed ast-ksh.2010-03-09
		(
			cmd='@@TYPE@@ -a @@VAR@@=( [1][3]=90 ) ;          function x { nameref nz=$1 ;               print " $(( round(nz) ))==$(( round($nz) ))" ; } ; x @@VAR@@[1][3]'	; stdoutpattern=' 90==90'
		)
		(
			cmd='@@TYPE@@ -a @@VAR@@ ;  @@VAR@@[1][3][5]=90 ; function x { nameref nz=$1 ;               print " $(( round(nz) ))==$(( round($nz) ))" ; } ; x @@VAR@@[1][3][5]'	; stdoutpattern=' 90==90'
		)
		# This crashed ast-ksh.2010-03-09
		(
			cmd='@@TYPE@@ -a @@VAR@@=( [1][3][5]=90 ) ;       function x { nameref nz=$1 ;               print " $(( round(nz) ))==$(( round($nz) ))" ; } ; x @@VAR@@[1][3][5]'	; stdoutpattern=' 90==90'
		)
		(
			cmd='@@TYPE@@ -a @@VAR@@ ;  @@VAR@@[1][3][5]=90 ; function x { nameref nz=${1}[$2][$3][$4] ; print " $(( round(nz) ))==$(( round($nz) ))" ; } ; x @@VAR@@ 1 3 5'	; stdoutpattern=' 90==90'
		)
		(
			cmd='@@TYPE@@ -A @@VAR@@ ;  @@VAR@@[1]=90 ;       function x { nameref nz=$1 ;               print " $(( round(nz) ))==$(( round($nz) ))" ; } ; x @@VAR@@[1]'		; stdoutpattern=' 90==90'
		)
		(
			cmd='@@TYPE@@ -A @@VAR@@=( [1]=90 ) ;             function x { nameref nz=$1 ;               print " $(( round(nz) ))==$(( round($nz) ))" ; } ; x @@VAR@@[1]'		; stdoutpattern=' 90==90'
		)
	)

	typeset testname
	integer i
	typeset mode
	typeset cmd

	for (( i=0 ; i < ${#tests[@]} ; i++ )) ; do
		for ty in \
			'typeset' \
			'integer' \
			'float' \
			'typeset -i' \
			'typeset -si' \
			'typeset -li' \
			'typeset -E' \
			'typeset -F' \
			'typeset -X' \
			'typeset -lE' \
			'typeset -lF' \
			'typeset -lX' ; do
			for mode in \
				'plain' \
				'in_compound' \
				'in_indexed_compound_array' \
				'in_2d_indexed_compound_array' \
				'in_4d_indexed_compound_array' \
				'in_associative_compound_array' \
				'in_compound_nameref' \
				'in_indexed_compound_array_nameref' \
				'in_2d_indexed_compound_array_nameref' \
				'in_4d_indexed_compound_array_nameref' \
				'in_associative_compound_array_nameref' \
				 ; do
				nameref tst=tests[i]
			
				cmd="${tst.cmd//@@TYPE@@/${ty}}"
				
				case "${mode}" in
					'plain')
						cmd="${cmd//@@VAR@@/z}"
						;;
					# compound
					'in_compound')
						cmd="compound c ; ${cmd//@@VAR@@/c.z}"
						;;
					'in_indexed_compound_array')
						cmd="compound -a c ; ${cmd//@@VAR@@/c[11].z}"
						;;
					'in_2d_indexed_compound_array')
						cmd="compound -a c ; ${cmd//@@VAR@@/c[17][19].z}"
						;;
					'in_4d_indexed_compound_array')
						cmd="compound -a c ; ${cmd//@@VAR@@/c[17][19][23][27].z}"
						;;
					'in_associative_compound_array')
						cmd="compound -A c ; ${cmd//@@VAR@@/c[info].z}"
						;;
					# nameref
					'in_compound_nameref')
						cmd="compound c ; nameref ncr=c.z ; ${cmd//@@VAR@@/ncr}"
						;;
					'in_indexed_compound_array_nameref')
						cmd="compound -a c ; nameref ncr=c[11].z ; ${cmd//@@VAR@@/ncr}"
						;;
					'in_2d_indexed_compound_array_nameref')
						cmd="compound -a c ; nameref ncr=c[17][19].z ; ${cmd//@@VAR@@/ncr}"
						;;
					'in_4d_indexed_compound_array_nameref')
						cmd="compound -a c ; nameref ncr=c[17][19][23][27].z ; ${cmd//@@VAR@@/ncr}"
						;;
					'in_associative_compound_array_nameref')
						cmd="compound -A c ; nameref ncr=c[info].z ; ${cmd//@@VAR@@/ncr}"
						;;
					*)
						err_exit "Unexpected mode ${mode}"
						;;
				esac
				
				testname="${0}/${cmd}"

				out.stderr="${ { out.stdout="${ ${SHELL} -o nounset -o errexit -c "${cmd}" ; (( out.res=$? )) ; }" ; } 2>&1 ; }"

			        [[ "${out.stdout}" == ${tst.stdoutpattern}      ]] || err_exit "${testname}: Expected stdout to match $(printf '%q\n' "${tst.stdoutpattern}"), got $(printf '%q\n' "${out.stdout}")"
       				[[ "${out.stderr}" == ''			]] || err_exit "${testname}: Expected empty stderr, got $(printf '%q\n' "${out.stderr}")"
				(( out.res == 0 )) || err_exit "${testname}: Unexpected exit code ${out.res}"
			done
		done
	done
	
	return 0
}

function test_has_iszero
{
	typeset str
	integer i
	
	typeset -r -a tests=(
		'(( iszero(0)   )) && print "OK"'
		'(( iszero(0.)  )) && print "OK"'
		'(( iszero(-0)  )) && print "OK"'
		'(( iszero(-0.) )) && print "OK"'
		'float n=0.  ; (( iszero(n) )) && print "OK"'
		'float n=+0. ; (( iszero(n) )) && print "OK"'
		'float n=-0. ; (( iszero(n) )) && print "OK"'
		'float n=1.  ; (( iszero(n) )) || print "OK"'
		'float n=1.  ; (( iszero(n-1.) )) && print "OK"'
		'float n=-1. ; (( iszero(n+1.) )) && print "OK"'
	)
	
	for (( i=0 ; i < ${#tests[@]} ; i++ )) ; do
		str="$( ${SHELL} -o errexit -c "${tests[i]}" 2>&1 )" || err_exit "test $i: returned non-zero exit code $?"
		[[ "${str}" == 'OK' ]] || err_exit "test $i: expected 'OK', got '${str}'"
	done

	return 0
}


# test deeply tested arithmetric expressions.
# ast-ksh.2010-03-10 crashed like this with this test:
# core 'core' of 27127:   ksh93 -c print -- $(( asinh(acosh(atanh(sin(cos(tan(atan(acos(asin(tan
#  2700000000000000 ???????? (1001095e0, 2f, 0, ffffffff7fffd708, ffffffff7d5a1235, ffffffff7fffd656)
#  ffffffff7d559050 comsubst (100102510, 0, 3ffe0acd00fe63b9, 6ca357895761ae66, 1001095c0, ffffffff7d6ab7a0) + 410
#  ffffffff7d5562fc varsub (100102510, 1b, 0, ffffffff7d6ab718, 1, ffffffff7d6a8000) + 4f4
#  ffffffff7d554980 copyto (100102510, 0, 1, ffffffff7e2f189c, 0, ffffffff7d6a8000) + 8dc
#  ffffffff7d5537ac sh_macexpand (ffffffff7d6aaf28, 100109440, ffffffff7fffe488, 37a0, 3400, 0) + 234
#  ffffffff7d538368 sh_argbuild (1001093e0, ffffffff7ffff35c, 100109440, 0, ffffffff7d6aaf28, 800) + b8
#  ffffffff7d582480 sh_exec (1001093e0, 25, 0, 0, 1, 0) + 3d0
#  ffffffff7d55c874 exfile (24, 25, 1001093e0, 24, 4, 4) + c1c
#  ffffffff7d55bc3c sh_main (0, 40000200001, 0, ffffffff7d6a8000, ffffffff7d6aaf28, ffffffff7d6a8000) + ca0
#  0000000100000d58 main (3, ffffffff7ffffa28, ffffffff7ffffa48, 1f5940, 1c80, 10000) + 44
#  0000000100000cfc _start (0, 0, 0, 0, 0, 0) + 17c
function test_deeply_nested_arithmetric_expression
{
	typeset out
	integer res
	
	out=${ ${SHELL} -c 'print -- $(( asinh(acosh(atanh(sin(cos(tan(atan(acos(asin(tanh(cosh(sinh(asinh(acosh(atanh(sin(cos(tan(atan(acos(asin(tanh(cosh(sinh(.5)))))))))))))))))))))))) ))' 2>&1 ; (( res=$? )) ; }

	[[ "${out}" == '0.5' ]] || err_exit "Expected output==0.5, got ${ printf '%q\n' "${out}" ; }"
	(( res == 0 )) || err_exit "Unexpected exit code $?"

	return 0
}

# ast-ksh.2010-06-17 and older do not preserve the type of a math function's
# return value, they fail the test below with a "Division by zero" error
# because they treat the right operand of the division as integer and not
# as floating point.
function test_return_type_of_math_function
{
	typeset out
	integer res
	
	out=${ ${SHELL} -c 'print -- $(( 1 / sin(4)))' 2>&1 ; (( res=$? )) ; }

	[[ "${out}" == -1.3213487088* ]] || err_exit "Expected output to match -1.3213487088*, got ${ printf '%q\n' "${out}" ; }"
	(( res == 0 )) || err_exit "Unexpected exit code $?"

	return 0
}

# test whether floating point implements minimum fraction bit length
# for IEEE754-2008 floating point values
# This is required to gurantee that scientific applications work as expected
function test_fraction_precision
{
	typeset out
	integer res
	typeset ti # test index
	typeset -r pi_str='0x3.243F6A8885A308D313198A2E03707344A4093822299F31D0082EFA98EC4Ep+00' # PI in C99 hexfloat
	
	# algorithm: tan(pi/2.) should return Infinitive but for real-world
	# IEEE754 floating point datatypes the storage for the fraction part
	# of the value is limited. tan(pi/2.) therefore returns the maximum
	# representable value (positive or negative) and log2() will return
	# the number of bits (plus one).
	compound -A tests=(
		[E]=(
			cmd="typeset -E pi=${pi_str} ; print -- \$(( int(log2(fabs(tan(pi/2.)))) ))"
			integer fraction_minprec=51
			integer fraction_prec=-1
		)
		[lE]=(
			cmd="typeset -lE pi=${pi_str} ; print -- \$(( int(log2(fabs(tan(pi/2.)))) ))"
			integer fraction_minprec=${ [[ "${ uname -p ; }" == 'sparc' ]] && print '112' || print '63' ; }
			integer fraction_prec=-1
		)
		[X]=(
			cmd="typeset -X pi=${pi_str} ; print -- \$(( int(log2(fabs(tan(pi/2.)))) ))"
			integer fraction_minprec=51
			integer fraction_prec=-1
		)
		[lX]=(
			cmd="typeset -lX pi=${pi_str} ; print -- \$(( int(log2(fabs(tan(pi/2.)))) ))"
			integer fraction_minprec=${ [[ "${ uname -p ; }" == 'sparc' ]] && print '112' || print '63' ; }
			integer fraction_prec=-1
		)
	)
	
	for ti in "${!tests[@]}" ; do
		nameref t=tests[${ti}]
	
		out=${ ${SHELL} -c "${t.cmd}" 2>&1 ; (( res=$? )) ; }

		[[ "${out}" == [[:digit:]]* ]] || err_exit "${ti}: Expected output to match [[:digit:]]*, got ${ printf '%q\n' "${out}" ; }"
		(( res == 0 )) || err_exit "${ti}: Unexpected exit code $?"
	
		if [[ "${out}" == ~(Elr)[[:digit:]]+ ]] ; then
			t.fraction_prec="${out}"
		fi
		
		(( t.fraction_prec >= t.fraction_minprec )) || err_exit "${ti}: fraction precision lower (${t.fraction_prec}) than expected (${t.fraction_minprec})"
	done
	
	(( tests[lE].fraction_prec > tests[E].fraction_prec )) || err_exit "${ti}: fraction precision for typeset -lE (${tests[lE].fraction_prec}) should be larger than for typeset -E (${tests[lE].fraction_prec}))"
	(( tests[lX].fraction_prec > tests[X].fraction_prec )) || err_exit "${ti}: fraction precision for typeset -lX (${tests[lX].fraction_prec}) should be larger than for typeset -X (${tests[lX].fraction_prec}))"

	return 0
}

function test_short_integer
{
        compound out=( typeset stdout stderr ; integer res )
	compound -r -a tests=(
		( cmd='integer -s -r -a x=( 1 2 3 ) ; print "${x[2]}"' stdoutpattern='3' )
		( cmd='integer -s -r -A x=( [0]=1 [1]=2 [2]=3 ) ; print "${x[2]}"' stdoutpattern='3' )
		# 2D integer arrays: the following two tests crash for both "integer -s" and "integer"
		( cmd='integer    -r -a x=( [0]=( [0]=1 [1]=2 [2]=3 ) [1]=( [0]=4 [1]=5 [2]=6 ) [2]=( [0]=7 [1]=8 [2]=9 ) ) ; print "${x[1][1]}"' stdoutpattern='5' )
		( cmd='integer -s -r -a x=( [0]=( [0]=1 [1]=2 [2]=3 ) [1]=( [0]=4 [1]=5 [2]=6 ) [2]=( [0]=7 [1]=8 [2]=9 ) ) ; print "${x[1][1]}"' stdoutpattern='5' )
   	)
	typeset testname
	integer i

	for (( i=0 ; i < ${#tests[@]} ; i++ )) ; do
		nameref tst=tests[i]
		testname="${0}/${i}"

		out.stderr="${ { out.stdout="${ ${SHELL} -o nounset -o errexit -c "${tst.cmd}" ; (( out.res=$? )) ; }" ; } 2>&1 ; }"

	        [[ "${out.stdout}" == ${tst.stdoutpattern}      ]] || err_exit "${testname}: Expected stdout to match $(printf '%q\n' "${tst.stdoutpattern}"), got $(printf '%q\n' "${out.stdout}")"
       		[[ "${out.stderr}" == ''			]] || err_exit "${testname}: Expected empty stderr, got $(printf '%q\n' "${out.stderr}")"
		(( out.res == 0 )) || err_exit "${testname}: Unexpected exit code ${out.res}"
	done
	
	return 0
}

# run tests
test_arithmetric_expression_accesss_array_element_through_nameref
test_has_iszero
test_deeply_nested_arithmetric_expression
test_return_type_of_math_function
test_fraction_precision
test_short_integer

cd "${ocwd}"
rmdir "${tmpdir}" || err_exit "Cannot remove temporary directory ${tmpdir}."

# tests done
exit $((Errors<125?Errors:125))
